#!/usr/bin/env python
"""
_Tier0Plugin_

Plugin which collects information from the agent and updates
the state of the current tier-0 workflows in the system.

Created on Nov 2, 2012

@author: dballest
"""

from builtins import filter

import re
import threading
import traceback

from WMComponent.AnalyticsDataCollector.Plugins.PluginInterface import PluginInterface
from WMCore.DAOFactory import DAOFactory
from WMCore.WMException import WMException
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper


def getTier0Regex():
    """
    _getTier0Regex_

    Define the Tier0 states and matching regex out of the object
    so it can be fetched without instancing the plugin.
    This are the uncompiled regular expressions in the correct
    order of the states
    """
    regexDict = {'Repack': [('Merge', [r'^/Repack_Run[0-9]+_Stream[\w]+/Repack$']),
                            ('Processing Done', [r'^/Repack_Run[0-9]+_Stream[\w]+/Repack/RepackMerge[\w]+$'])],

                 'PromptReco': [('AlcaSkim', [r'^/PromptReco_Run[0-9]+_[\w]+/Reco$']),
                                ('Merge', [r'^/PromptReco_Run[0-9]+_[\w]+/Reco/AlcaSkim$']),
                                ('Harvesting', [r'^/PromptReco_Run[0-9]+_[\w]+/Reco/AlcaSkim/AlcaSkimMerge[\w]+$',
                                                r'^/PromptReco_Run[0-9]+_[\w]+/Reco/RecoMerge[\w]+$']),
                                ('Processing Done',
                                 [r'^/PromptReco_Run[0-9]+_[\w]+/Reco/RecoMerge[\w]+/RecoMerge[\w]+DQMHarvest[\w]+$'])
                                ],

                 'Express': [('Merge', [r'^/Express_Run[0-9]+_Stream[\w]+/Express$']),
                             ('Harvesting', [r'^/Express_Run[0-9]+_Stream[\w]+/Express/ExpressMerge[\w]+$',
                                             r'^/Express_Run[0-9]+_Stream[\w]+/Express/ExpressAlcaSkim[\w]+$']),
                             ('Processing Done', [
                                 r'^/Express_Run[0-9]+_Stream[\w]+/Express/ExpressMerge[\w]+/ExpressMerge[\w]+DQMHarvest[\w]+$',
                                 r'^/Express_Run[0-9]+_Stream[\w]+/Express/ExpressAlcaSkim[\w]+/ExpressAlcaSkim[\w]+AlcaHarvest[\w]+$'])
                             ]}
    return regexDict


class Tier0PluginError(WMException):
    """
    _Tier0PluginError_

    An error in the Tier0 plugin
    """

    def __init__(self, msg):
        """
        __init__

        Initialize the error, just accepts a message without error code
        """
        WMException.__init__(self, msg)
        return


class Tier0Plugin(PluginInterface):
    """
    _Tier0Plugin_

    Tier0 plugin main class
    """

    def __init__(self):
        """
        __init__

        Initialize the plugin object
        """
        PluginInterface.__init__(self)

        self.myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package='WMCore.WMBS',
                                     logger=self.myThread.logger,
                                     dbinterface=self.myThread.dbi)

        self.logger = self.myThread.logger

        # To get finished subscriptions it needs the cleanout state index
        getCleanoutState = self.daoFactory(classname='Jobs.GetStateID')
        self.cleanoutState = getCleanoutState.execute(state='cleanout')

        # Load the DAOs
        self.getFinishedTasks = self.daoFactory(classname='Workflow.GetFinishedTasks')
        self.getFinishedTasksNoInjection = self.daoFactory(classname='Subscriptions.GetSemiFinishedTasks')

        # To avoid doing many I/O operations, cache the list of tasks for each workflow
        self.taskCache = {}

        self.setupStateRegex()

        return

    def setupStateRegex(self):
        """
        _setupStateRegex_

        Each state is mapped to a regular expression in the taskpaths
        this depends on the current implementation of the T0. The regular
        expressions obtained from the function Tier0Plugin.getTier0Regex
        are compiled and stored in a dictionary.

        The structure of the dictionary is as follows:

        { <workflowType> : [(<state>, [<matchObject>, <matchObject>]), ...],
          ...
        }
        """
        rawRegex = getTier0Regex()

        self.tier0Regex = {}

        for workflowType in rawRegex:
            self.tier0Regex[workflowType] = []
            for pair in rawRegex[workflowType]:
                compiledRegex = (pair[0], [re.compile(x).match for x in pair[1]])
                self.tier0Regex[workflowType].append(compiledRegex)
        return

    def __call__(self, requestDocs, localSummaryCouchDB, centralRequestCouchDB):
        """
        __call__

        Call to this plugin, this method takes care of executing
        the logic of the plugin, see guidelines in parent class documentation.
        """
        # Get all current finished tasks (i.e. all associated subscriptions are finished)
        finishedTasks = self.getFinishedTasks.execute()
        self.logger.debug("Found %d finished tasks" % len(finishedTasks))
        # Repack workflows are not marked injected and therefore its subscriptions
        # are not marked as finished until the injection happens (48h delay)
        # get those tasks with a looser criteria
        finishedTasksNoInjection = self.getFinishedTasksNoInjection.execute(state=self.cleanoutState,
                                                                            pattern='%Repack%')
        self.logger.debug("Found %d finished Repack tasks" % len(finishedTasksNoInjection))
        # Get workflows which are not closed yet or completed
        notClosedWorkflows = centralRequestCouchDB.getRequestByStatus(['new'])

        # Aggregate results by workflow name
        finishedTasks.extend(finishedTasksNoInjection)
        workflows = {}
        for entry in finishedTasks:
            workflowName = entry['name']
            if workflowName not in workflows:
                workflows[workflowName] = []
            if workflowName not in self.taskCache:
                self.loadTasks(workflowName, entry['spec'])
            workflows[workflowName].append(entry['task'])

        self.logger.info('Updating the status of %d workflows' % len(workflows))
        # Go through the reported workflows
        for workflowName in workflows:
            try:
                if workflowName in notClosedWorkflows:
                    # Nothing is completed for this workflow
                    continue

                # Sanity checks on workflow name
                tokens = workflowName.split('_')
                if not len(tokens):
                    self.logger.warning('This workflow does not match the Tier-0 naming structure, it will be ignored.')
                    continue
                workflowType = tokens[0]
                if workflowType not in self.tier0Regex:
                    self.logger.warning('This workflow does not match the Tier-0 naming structure, it will be ignored.')
                    continue

                completedTaskList = workflows[workflowName]

                workflowStatus = self.determineCurrentStatus(workflowName, workflowType, completedTaskList)
                if workflowStatus is not None:
                    centralRequestCouchDB.updateRequestStatus(workflowName, workflowStatus)
            except Tier0PluginError as t0ex:
                # More specific exception, just log it anyway
                self.logger.error('Error ocurred while processing a doc:\n%s' % str(t0ex))
            except Exception as ex:
                # Plugins are meant to be not-critical
                # If something fails then just keep going
                self.logger.error('Error occurred while processing docs:\n%s' % str(ex))
                self.logger.error(traceback.format_exc())

        # Clean the task cache based on the documents we reported this cycle
        self.cleanTaskCache([x['workflow'] for x in requestDocs])

        return

    def loadTasks(self, workflowName, spec):
        """
        _loadTasks_

        Loads the list of tasks for the workflow,
        stores them in the cache if not present
        """
        if workflowName in self.taskCache:
            return
        try:
            workloadHelper = WMWorkloadHelper()
            workloadHelper.load(spec)
            tasks = workloadHelper.listAllTaskPathNames()
            self.taskCache[workflowName] = tasks
        except IOError as ex:
            msg = "Failed to load spec file %s\n" % spec
            msg += "Original IOError: %s" % str(ex)
            raise Tier0PluginError(msg)
        return

    def cleanTaskCache(self, reportedWorkflows):
        """
        _cleanTaskCache_

        Keep the memory footprint of this plugin smaller
        clean unused task lists
        """
        self.logger.debug('Cleaning up task cache')
        for workflow in list(self.taskCache):
            if workflow not in reportedWorkflows:
                self.taskCache.pop(workflow)
        return

    def determineCurrentStatus(self, workflowName, workflowType, completedTasks):
        """
        _determineCurrentStatus_

        Process a completed task list for a workflow and
        get the most advanced status of the workflow.
        """
        self.logger.debug('Calculating status for %s' % workflowName)
        currentStatus = None
        typeRegex = self.tier0Regex[workflowType]
        fullTaskList = self.taskCache[workflowName]
        for pair in typeRegex:
            completedTasksForStatus = 0
            totalTasksForStatus = 0
            for regex in pair[1]:
                completedTasksForStatus += len(list(filter(regex, completedTasks)))
                totalTasksForStatus += len(list(filter(regex, fullTaskList)))
            if completedTasksForStatus == totalTasksForStatus:
                currentStatus = pair[0]
            else:
                break
        self.logger.debug('Status is %s' % currentStatus)
        self.logger.debug('Completed task list: %s' % str(completedTasksForStatus))

        return currentStatus



#!/usr/bin/env python
"""
_AssociateWorkflowToFile_

MySQL implementation of DBS3Buffer.DBSBufferFiles.AssociateWorkflowToFile

Created on Oct 12, 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class AssociateWorkflowToFile(DBFormatter):
    """
    _AssociateWorkflowToFile_

    Associate the given workflow and task to an existing dbsbuffer file,
    the lfn of an existing file in dbsbuffer is the input
    """

    sql = """UPDATE dbsbuffer_file
             SET workflow = (SELECT id FROM dbsbuffer_workflow
                             WHERE name = :name AND task = :task)
             WHERE lfn = :lfn
          """

    def execute(self, lfn, workflowName, task, conn = None, transaction = False):
        """
        _execute_

        Run the query
        """
        binds = {"lfn" : lfn, "name" : workflowName, "task" : task}

        self.dbi.processData(self.sql, binds,
                             conn = conn, transaction = transaction)

        return



"""
Thread used to monitor the agent draining process. Should
eventually report no issues meaning that the agent is ready
to be shutdown and a new version be put in place.
"""
from __future__ import division

from future.utils import viewitems

__all__ = []

import logging
import copy
from Utils.Timers import timeFunction
from WMComponent.AgentStatusWatcher.DrainStatusAPI import DrainStatusAPI
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.Services.ReqMgrAux.ReqMgrAux import ReqMgrAux
from WMCore.Services.PyCondor.PyCondorAPI import PyCondorAPI
from Utils.EmailAlert import EmailAlert

class DrainStatusPoller(BaseWorkerThread):
    """
    Collects information related to the agent drain status
    """
    # class variable that contains drain statistics
    drainStats = {}

    def __init__(self, config):
        """
        initialize properties specified from config
        """
        BaseWorkerThread.__init__(self)
        self.config = config
        self.drainAPI = DrainStatusAPI(config)
        self.condorAPI = PyCondorAPI()
        self.agentConfig = {}
        self.previousConfig = {}
        self.validSpeedDrainConfigKeys = ['CondorPriority', 'NoJobRetries', 'EnableAllSites']
        cacheduration = getattr(self.config.General, "ReqMgrAuxCacheDuration", 5 / 60)  # 5 minutes
        self.reqAuxDB = ReqMgrAux(self.config.General.ReqMgr2ServiceURL, httpDict={'cacheduration': cacheduration})
        self.emailAlert = EmailAlert(config.EmailAlert.dictionary_())
        self.condorStates = ("Running", "Idle")

    @timeFunction
    def algorithm(self, parameters):
        """
        Update drainStats if agent is in drain mode
        """
        logging.info("Running agent drain algorithm...")
        if self.agentConfig:
            # make a copy of the previous agent aux db configuration to compare against later
            self.previousConfig = copy.deepcopy(self.agentConfig)
        # grab a new copy of the agent aux db configuration
        self.agentConfig = self.reqAuxDB.getWMAgentConfig(self.config.Agent.hostName)
        if not self.agentConfig:
            logging.error("Failed to fetch agent configuration from the auxiliary DB")
            return

        try:
            # see if the agent is in drain mode
            if self.agentConfig["UserDrainMode"] or self.agentConfig["AgentDrainMode"]:
                # check to see if the agent hit any speed drain thresholds
                thresholdsHit = self.checkSpeedDrainThresholds()
                if thresholdsHit:
                    logging.info("Updating agent configuration for speed drain...")
                    self.updateAgentSpeedDrainConfig(thresholdsHit)
                # now collect drain statistics
                DrainStatusPoller.drainStats = self.drainAPI.collectDrainInfo()
                logging.info("Finished collecting agent drain status.")
                logging.info("Drain stats: %s", str(DrainStatusPoller.drainStats))
            else:
                logging.info("Agent not in drain mode. Resetting flags and skipping drain check...")
                self.resetAgentSpeedDrainConfig()

            # finally, check for any changes in drain status
            self.checkDrainStatusChanges()

        except Exception as ex:
            msg = "Error occurred, will retry later:\n"
            msg += str(ex)
            logging.exception(msg)

    @classmethod
    def getDrainInfo(cls):
        """
        Return drainStats class variable
        """
        return cls.drainStats

    def checkDrainStatusChanges(self):
        """
        Check to see if any drain statuses have changed in the auxiliary db
        If yes, send email notification and update local drain thread variables

        """
        message = ""
        drainStatusKeys = ['UserDrainMode', 'AgentDrainMode', 'SpeedDrainMode']

        if not self.previousConfig:
            return

        for key in drainStatusKeys:
            if self.previousConfig[key] != self.agentConfig[key]:
                message += "Agent had a drain status transition to %s = %s\n" % (str(key), str(self.agentConfig[key]))

        if message:
            self.emailAlert.send("DrainMode status change on " + getattr(self.config.Agent, "hostName"), message)
            logging.info("Drain mode status change: %s", message)

        return

    def updateAgentSpeedDrainConfig(self, thresholdsHit):
        """
        Takes a list of speed drain configuration keys and updates the agent configuration
        """
        updateConfig = False
        condorPriorityFlag = False
        speedDrainConfig = self.agentConfig.get("SpeedDrainConfig")

        if 'CondorPriority' in thresholdsHit:
            logging.info("Bumping condor job priority to 999999 for Production/Processing pending jobs.")
            self.condorAPI.editCondorJobs(
                "JobStatus=?=1 && (CMS_JobType =?= \"Production\" || CMS_JobType =?= \"Processing\")",
                "JobPrio", "999999")
            condorPriorityFlag = True

        if condorPriorityFlag != speedDrainConfig['CondorPriority']['Enabled']:
            # CondorPriority setting is irreversible so the flag only indicates weather
            # priority is increased or not. It is not checked by other components
            logging.info("Enabling CondorPriority flag.")
            speedDrainConfig['CondorPriority']['Enabled'] = condorPriorityFlag
            updateConfig = True

        if 'NoJobRetries' in thresholdsHit:
            logging.info("Enabling NoJobRetries flag: Error Handler won't retry the jobs")
            # ErrorHandler will pick this up and set max retries to 0
            speedDrainConfig['NoJobRetries']['Enabled'] = True
            updateConfig = True

        if 'EnableAllSites' in thresholdsHit:
            logging.info("Enabling EnableAllSites flag: Updating agent to submit to all sites.")
            # setting this value to True makes JobSubmitterPoller ignore site status
            speedDrainConfig['EnableAllSites']['Enabled'] = True
            updateConfig = True

        # update the aux db speed drain config with any changes
        if updateConfig:
            self.agentConfig['SpeedDrainMode'] = True
            self.reqAuxDB.updateWMAgentConfig(self.config.Agent.hostName, self.agentConfig)

        return

    def resetAgentSpeedDrainConfig(self):
        """
        resetting SpeedDrainMode to False and SpeedDrainConfig Enabled to False
        """

        if self.agentConfig.get("SpeedDrainMode"):
            self.agentConfig['SpeedDrainMode'] = False
            speedDrainConfig = self.agentConfig.get("SpeedDrainConfig")
            for key, v in viewitems(speedDrainConfig):
                if key in self.validSpeedDrainConfigKeys and v['Enabled']:
                    speedDrainConfig[key]['Enabled'] = False

            self.reqAuxDB.updateWMAgentConfig(self.config.Agent.hostName, self.agentConfig)
        return

    def checkSpeedDrainThresholds(self):
        """
        Check the current number of jobs in Condor and create a list of agent configuration parameters
        that need updated for speed draining
        """
        enableKeys = []
        # first, update our summary of condor jobs
        totalJobs = self.getTotalCondorJobs()
        if totalJobs is None:
            msg = "Cannot check speed drain because there was an error fetching job summary from HTCondor."
            msg += " Will retry again in the next cycle."
            logging.warning(msg)
            return []

        # get the current speed drain status
        speedDrainConfig = self.agentConfig.get("SpeedDrainConfig")

        # loop through the speed drain configuration and make a list of what thresholds have been hit
        for k, v in viewitems(speedDrainConfig):
            # make sure keys in the speed drain config are valid
            if k in self.validSpeedDrainConfigKeys and isinstance(v['Threshold'], int) and isinstance(v['Enabled'], bool):
                # we always want to apply the condor priority change if the threshold is hit
                if not v['Enabled'] or k == 'CondorPriority':
                    logging.info("Checking speed drain threshold for %s. ", k)
                    if totalJobs < v['Threshold']:
                        logging.info("Agent will update speed drain configuration for %s. ", k)
                        enableKeys.append(k)
            else:
                logging.warning("Speed drain configuration error for %s.  Please check aux db contents.", k)

        return enableKeys

    def getTotalCondorJobs(self):
        """
        Retrieve a summary of the jobs in condor and return an absolute number
        of the jobs in Idle and Running states.
        :return: returns an integer with the total number of jobs, or None if it failed.
        """
        jobs = self.condorAPI.getCondorJobsSummary()
        if not jobs:
            return None

        results = 0
        if jobs:
            for state in self.condorStates:
                results += int(jobs[0].get(state))
        return results



#!/usr/bin/env python
"""
_Delete_

MySQL implementation of JobGroup.Delete

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Delete(DBFormatter):
    sql = "delete from wmbs_jobgroup where id = :id"

    def getBinds(self, id):
        return self.dbi.buildbinds(self.dbi.makelist(id), 'id')

    def execute(self, id, conn = None, transaction = False):
        self.dbi.processData(self.sql, self.getBinds(id),
                         conn = conn, transaction = transaction)
        return True #or raise



"""
_WorkUnit.ExistsByID_

Oracle implementation of WorkUnit.ExistsByID
"""

from __future__ import absolute_import, division, print_function

from WMCore.WMBS.MySQL.WorkUnit.ExistsByID import ExistsByID as ExistsByIDMySQL


class ExistsByID(ExistsByIDMySQL):
    """
    _WorkUnit.ExistsByID_

    Oracle implementation of WorkUnit.ExistsByID
    """

    pass
